#!/usr/bin/env python
#
#   Copyright 2019 Andrea Bonomi <andrea.bonomi@gmail.com>
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the Licens
#
__author__ = 'Andrea Bonomi <andrea.bonomi@gmail.com>'
__version__ = '2.4.0'

from flask import Blueprint
from airflow.plugins_manager import AirflowPlugin
from airflow_code_editor.commons import STATIC
from airflow_code_editor.flask_admin_view import admin_view
from airflow_code_editor.app_builder_view import appbuilder_view

__all__ = [
    'CodeEditorPlugin'
]


code_editor_plugin_blueprint = Blueprint(
    'code_editor_plugin_blueprint',
    __name__,
    template_folder='templates',
    static_folder='static',
    static_url_path=STATIC
)


# Plugin
class CodeEditorPlugin(AirflowPlugin):
    name = 'editor_plugin'
    operators = []
    flask_blueprints = [code_editor_plugin_blueprint]
    hooks = []
    executors = []
    admin_views = [admin_view]
    menu_links = []
    appbuilder_views = [appbuilder_view]



#!/usr/bin/env python

import airflow
from functools import wraps
try:
    from flask_appbuilder import has_access
except (ImportError, ModuleNotFoundError):
    def has_access(x):
        return x

__all__ = [
    'login_required',
    'has_access'
]


def login_required(func):
    # when airflow loads plugins, login is still None.
    @wraps(func)
    def func_wrapper(*args, **kwargs):
        if airflow.login:
            return airflow.login.login_required(func)(*args, **kwargs)
        return func(*args, **kwargs)
    return func_wrapper



#!/usr/bin/env python
#
#   Copyright 2019 Andrea Bonomi <andrea.bonomi@gmail.com>
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the Licens
#

import os
import os.path
import logging
import mimetypes
from flask import (
    abort,
    jsonify,
    request,
    send_file
)
from airflow_code_editor.commons import (
    HTTP_404_NOT_FOUND
)
from airflow_code_editor.utils import (
    git_absolute_path,
    execute_git_command,
    normalize_path
)

__all__ = [
    'AbstractCodeEditorView'
]


class AbstractCodeEditorView(object):

    def _index(self, session=None):
        return self._render('index')

    def _save(self, session=None, path=None):
        try:
            data = request.form['data']
            # Newline fix (remove cr)
            data = data.replace('\r', '').rstrip()
            fullpath = git_absolute_path(path)
            os.makedirs(os.path.dirname(fullpath), exist_ok=True)
            with open(fullpath, 'w') as f:
                f.write(data)
                f.write('\n')
            return jsonify({
                'path': normalize_path(path)
            })
        except Exception as ex:
            logging.error(ex)
            if hasattr(ex, 'strerror'):
                message = ex.strerror
            elif hasattr(ex, 'message'):
                message = ex.message
            else:
                message = str(ex)
            return jsonify({
                'path': normalize_path(path),
                'error': {
                    'message': 'Error saving {path}: {message}'.format(path=path, message=message)
                }
            })

    def _git_repo(self, session, path):
        if request.method == 'POST':
            return self._git_repo_post(session, path)
        else:
            return self._git_repo_get(session, path)

    def _git_repo_get(self, session, path):
        " Get a file from GIT (invoked by the HTTP GET method) "
        return execute_git_command(["cat-file", "-p", path])

    def _git_repo_post(self, session, path):
        " Execute a GIT command (invoked by the HTTP POST method) "
        git_args = request.form.getlist('args[]')
        return execute_git_command(git_args)

    def _load(self, session, path):
        " Send the contents of a file to the client "
        try:
            path = normalize_path(path)
            if path.startswith('~git/'):
                # Download git blob - path = '~git/<hash>/<name>'
                _, path, filename = path.split('/', 3)
                response = execute_git_command(["cat-file", "-p", path])
                response.headers['Content-Disposition'] = 'attachment; filename="%s"' % filename
                try:
                    content_type = mimetypes.guess_type(filename)[0]
                    if content_type:
                        response.headers['Content-Type'] = content_type
                except Exception:
                    pass
                return response
            else:
                # Download file
                fullpath = git_absolute_path(path)
                return send_file(fullpath, as_attachment=True)
        except Exception as ex:
            logging.error(ex)
            abort(HTTP_404_NOT_FOUND)



#!/usr/bin/env python
#
#   Copyright 2019 Andrea Bonomi <andrea.bonomi@gmail.com>
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the Licens

import os
import os.path
import logging
import subprocess
import threading
import shlex
from datetime import datetime
from collections import namedtuple
from flask import make_response
from flask_login import current_user
from airflow import configuration
from airflow_code_editor.commons import (
    HTTP_200_OK,
    HTTP_404_NOT_FOUND,
    SUPPORTED_GIT_COMMANDS,
    PLUGIN_NAME,
    PLUGIN_DEFAULT_CONFIG,
    ROOT_MOUNTPOUNT
)

__all__ = [
    'normalize_path',
    'get_plugin_config',
    'get_plugin_boolean_config',
    'get_root_folder',
    'git_absolute_path',
    'execute_git_command'
]


def normalize_path(path):
    comps = (path or '/').split('/')
    result = []
    for comp in comps:
        if comp in ('', '.'):
            pass
        elif (comp != '..' or (result and result[-1] == '..')):
            result.append(comp)
        elif result:
            result.pop()
    return '/'.join(result)


def get_plugin_config(key):
    " Get a plugin configuration/default for a given key "
    return configuration.conf.get(PLUGIN_NAME, key, fallback=PLUGIN_DEFAULT_CONFIG[key])


def get_plugin_boolean_config(key):
    " Get a plugin boolean configuration/default for a given key "
    return configuration.conf.getboolean(PLUGIN_NAME, key, fallback=PLUGIN_DEFAULT_CONFIG[key])


def prepare_response(git_cmd, result=None, stderr=None, returncode=0):
    if result is None:
        result = stderr
    elif stderr:
        result = result + stderr
    if git_cmd == 'cat-file':
        response = make_response(result, HTTP_200_OK if returncode == 0 else HTTP_404_NOT_FOUND)
        response.headers['Content-Type'] = 'text/plain'
    else:
        response = make_response(result)
        response.headers['X-Git-Return-Code'] = str(returncode)
        response.headers['X-Git-Stderr-Length'] = str(len(stderr or ''))
    return response


def prepare_git_env():
    " Prepare the environ for git "
    env = dict(os.environ)
    git_author_name = get_plugin_config('git_author_name')
    if not git_author_name:
        try:
            git_author_name = '%s %s' % (current_user.first_name, current_user.last_name)
        except Exception:
            pass
    if git_author_name:
        env['GIT_AUTHOR_NAME'] = git_author_name
        env['GIT_COMMITTER_NAME'] = git_author_name
    git_author_email = get_plugin_config('git_author_email')
    if not git_author_email:
        try:
            git_author_email = current_user.email
        except Exception:
            pass
    if git_author_email:
        env['GIT_AUTHOR_EMAIL'] = git_author_email
        env['GIT_COMMITTER_EMAIL'] = git_author_email
    return env


def get_root_folder():
    " Return the configured root folder or Airflow DAGs folder "
    return os.path.abspath(get_plugin_config('root_directory') or configuration.conf.get('core', 'dags_folder'))


def git_absolute_path(git_path):
    " Git relative path to absolute path "
    path = normalize_path(git_path)
    if path.startswith('~'):
        # Expand paths beginning with '~'
        prefix, remain = path.split('/', 1) if '/' in path else (path, '')
        try:
            return os.path.join(mount_points[prefix[1:]].path, remain)
        except KeyError:
            pass
    return os.path.join(get_root_folder(), path)


_execute_git_command_lock = threading.Lock()


def execute_git_command(git_args):
    with _execute_git_command_lock:
        logging.info(' '.join(git_args))
        git_cmd = git_args[0] if git_args else None
        try:
            cwd = get_root_folder()
            if not os.path.exists(os.path.join(cwd, '.git')) and get_plugin_boolean_config('git_init_repo'):
                init_git_repo()
            if git_cmd == 'ls-local':
                stdout = git_ls_local(git_args)
                stderr = None
                returncode = 0
            elif git_cmd == 'mounts':
                stdout = git_mounts(git_args)
                stderr = None
                returncode = 0
            elif git_cmd in SUPPORTED_GIT_COMMANDS:
                git_default_args = shlex.split(get_plugin_config('git_default_args'))
                cmd = [get_plugin_config('git_cmd')] + git_default_args + git_args
                git = subprocess.Popen(cmd,
                                       stdin=subprocess.PIPE,
                                       stdout=subprocess.PIPE,
                                       stderr=subprocess.PIPE,
                                       cwd=cwd,
                                       env=prepare_git_env())
                stdout, stderr = git.communicate()
                returncode = git.returncode
            else:
                stdout = None
                stderr = 'Command not supported: git %s' % ' '.join(git_args)
                returncode = 1
        except OSError as ex:
            logging.error(ex)
            stdout = None
            stderr = ex.strerror
            returncode = ex.errno
        except Exception as ex:
            logging.error(ex)
            stdout = None
            stderr = ex.message
            returncode = 1
        finally:
            return prepare_response(git_cmd, stdout, stderr, returncode)


def git_ls_local(git_args):
    " 'git ls-tree' like output for local folders "
    long_ = False
    if '-l' in git_args or '--long' in git_args:
        git_args = [arg for arg in git_args if arg not in ('-l', '--long')]
        long_ = True
    path = git_args[1] if len(git_args) > 1 else ''
    path = normalize_path(path.split('#', 1)[0])
    dirpath = git_absolute_path(path)
    result = []
    for name in sorted(os.listdir(dirpath)):
        if name.startswith('.') or name == '__pycache__':
            continue
        fullname = os.path.join(dirpath, name)
        s = os.stat(fullname)
        if os.path.isdir(fullname):
            type_ = 'tree'
            size_ = '-'
        else:
            type_ = 'blob'
            size_ = s.st_size
        relname = os.path.join('/', path, name)  # fullname[len(cwd):]
        if long_:
            mtime = datetime.utcfromtimestamp(s.st_mtime).isoformat()[:16]
            result.append('%06o %s %s#%s %8s\t%s' % (s.st_mode, type_, relname, mtime, size_, name))
        else:
            result.append('%06o %s %s\t%s' % (s.st_mode, type_, relname, name))
    return '\n'.join(result)


def git_mounts(git_args):
    " List mountpoints "
    return '\n'.join(sorted(k for k, v in mount_points.items() if not v.default))


def init_git_repo():
    " Initialize the git repository in root folder "
    cwd = get_root_folder()
    subprocess.call(['git', 'init', '.'], cwd=cwd)
    gitignore = os.path.join(cwd, '.gitignore')
    if not os.path.exists(gitignore):
        with open(gitignore, 'w') as f:
            f.write('__pycache__\n')
        subprocess.call(['git', 'add', '.gitignore'], cwd=cwd)
    subprocess.call(['git', 'commit', '-m', 'Initial commit'], cwd=cwd, env=prepare_git_env())


MountPoint = namedtuple('MountPoint', 'path default')


def read_mount_points_config():
    " Return the plugin configuration "
    global mount_points
    config = {
        ROOT_MOUNTPOUNT: MountPoint(path=get_root_folder(), default=True)
    }
    i = 0
    # Iterate over the configurations
    while True:
        suffix = str(i) if i != 0 else ''  # the first configuration doesn't have a suffix
        try:
            if not configuration.conf.has_option(PLUGIN_NAME, 'mount{}_name'.format(suffix)):
                break
        except Exception:  # backports.configparser.NoSectionError and friends
            break
        name = configuration.conf.get(PLUGIN_NAME, 'mount{}_name'.format(suffix))
        path = configuration.conf.get(PLUGIN_NAME, 'mount{}_path'.format(suffix))
        config[name] = MountPoint(path=path, default=False)
        i = i + 1
    mount_points = config


read_mount_points_config()



#!/usr/bin/env python
#
#   Copyright 2019 Andrea Bonomi <andrea.bonomi@gmail.com>
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the Licens
#

from flask_admin import BaseView, expose
import airflow
from airflow.utils.db import provide_session
from airflow_code_editor.code_editor_view import AbstractCodeEditorView
from airflow_code_editor.auth import login_required
from airflow_code_editor.commons import (
    ROUTE,
    MENU_CATEGORY,
    MENU_LABEL
)
assert(airflow)

__all__ = [
    'AdminCodeEditorView',
    'admin_view'
]


# ############################################################################
# Flask Admin

class AdminCodeEditorView(BaseView, AbstractCodeEditorView):

    @expose('/')
    @login_required
    @provide_session
    def index(self, session=None):
        return self._index(session)

    @expose('/repo', methods=['POST'])
    @login_required
    @provide_session
    def repo_base(self, session=None, path=None):
        return self._git_repo(session, path)

    @expose('/repo/<path:path>', methods=['GET', 'HEAD', 'POST'])
    @login_required
    @provide_session
    def repo(self, session=None, path=None):
        return self._git_repo(session, path)

    @expose('/files/<path:path>', methods=['POST'])
    @login_required
    @provide_session
    def save(self, session=None, path=None):
        return self._save(session, path)

    @expose('/files/<path:path>', methods=['GET'])
    @login_required
    @provide_session
    def load(self, session=None, path=None):
        return self._load(session, path)

    def _render(self, template, *args, **kargs):
        return self.render(template + '_admin.html',
                           airflow_refresh="airflow.refresh",
                           log_list='log.index_view',
                           *args, **kargs)


admin_view = AdminCodeEditorView(
    url=ROUTE,
    category=MENU_CATEGORY,
    name=MENU_LABEL
)



#!/usr/bin/env python
import os
import os.path
import re
import glob
from setuptools import (find_packages, setup)
from itertools import chain

VERSION_RE = re.compile('__version__\\s*=\\s*[\'"](.*)[\'"]')


def get_version():
    with open(os.path.join(os.path.dirname(os.path.abspath(__file__)),
                           'airflow_code_editor/airflow_code_editor.py')) as f:
        for line in f:
            match = VERSION_RE.match(line)
            if match:
                return match.group(1)
    raise Exception


def get_package_data():
    exts = ['js', 'css', 'html', 'svg', 'png', 'jpg', 'gif']
    files = list(chain(*[glob.glob('airflow_code_editor/**/*.%s' % x, recursive=True) for x in exts]))
    return [x.split('/', 1)[1] for x in files]


with open('README.md', 'r') as f:
    long_description = f.read()

with open('requirements.txt', 'r') as f:
    install_requires = f.read().split('\n')

setup(name='airflow_code_editor',
      version=get_version(),
      packages=find_packages(),
      include_package_data=True,
      package_data={'airflow_code_editor': get_package_data()},
      entry_points={
          'airflow.plugins': [
              'airflow_code_editor = airflow_code_editor.airflow_code_editor:CodeEditorPlugin'
          ]
      },
      zip_safe=False,
      url='https://github.com/andreax79/airflow-code-editor',
      author='Andrea Bonomi',
      author_email='andrea.bonomi@gmail.com',
      description='Apache Airflow code editor and file manager',
      long_description=long_description,
      long_description_content_type='text/markdown',
      install_requires=install_requires,
      license='Apache License, Version 2.0',
      classifiers=[
          'Operating System :: OS Independent',
          'Programming Language :: Python',
          'Topic :: Software Development :: Libraries :: Python Modules',
          'Programming Language :: Python :: 3',
          'Programming Language :: Python :: 3.3',
          'Programming Language :: Python :: 3.4',
          'Programming Language :: Python :: 3.5',
          'Programming Language :: Python :: 3.6',
          'Programming Language :: Python :: 3.7',
          'Programming Language :: Python :: 3.8',
      ])



#!/usr/bin/env python
#
#   Copyright 2019 Andrea Bonomi <andrea.bonomi@gmail.com>
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the Licens
#

__all__ = [
    'PLUGIN_NAME',
    'MENU_CATEGORY',
    'MENU_LABEL',
    'ROUTE',
    'STATIC',
    'CONFIG_SECTION',
    'SUPPORTED_GIT_COMMANDS',
    'HTTP_200_OK',
    'HTTP_404_NOT_FOUND',
    'PLUGIN_DEFAULT_CONFIG',
    'ROOT_MOUNTPOUNT'
]

PLUGIN_NAME = 'code_editor'
MENU_CATEGORY = 'Admin'
MENU_LABEL = 'DAGs Code Editor'
ROUTE = '/' + PLUGIN_NAME
STATIC = '/static/' + PLUGIN_NAME
CONFIG_SECTION = PLUGIN_NAME + '_plugin'
HTTP_200_OK = 200
HTTP_404_NOT_FOUND = 404
SUPPORTED_GIT_COMMANDS = [
    'add',
    'branch',
    'checkout',
    'cat-file',
    'commit',
    'diff',
    'log',
    'ls-tree',
    'reset',
    'rm',
    'show',
    'stage',
    'status',
    'tag',
    'unstage'
]
PLUGIN_DEFAULT_CONFIG = {
    'git_cmd': 'git',
    'git_default_args': '-c color.ui=true',
    'git_author_name': None,
    'git_author_email': None,
    'git_init_repo': True,
    'root_directory': None
}
ROOT_MOUNTPOUNT = 'root'



#!/usr/bin/env python
#
#   Copyright 2019 Andrea Bonomi <andrea.bonomi@gmail.com>
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the Licens
#

from flask_appbuilder import BaseView, expose
from airflow.utils.db import provide_session
from airflow.www_rbac.decorators import has_dag_access
from airflow_code_editor.auth import has_access
from airflow_code_editor.code_editor_view import AbstractCodeEditorView
from airflow_code_editor.commons import (
    ROUTE,
    MENU_CATEGORY,
    MENU_LABEL
)

__all__ = [
    'AppBuilderCodeEditorView',
    'appbuilder_view'
]


# ############################################################################
# AppBuilder (Airflow >= 1.10 and rbac = True)

class AppBuilderCodeEditorView(BaseView, AbstractCodeEditorView):
    route_base = ROUTE
    base_permissions = ['can_list']

    @expose('/')
    @has_dag_access(can_dag_edit=True)
    @has_access
    @provide_session
    def list(self, session=None):
        return self._index(session)

    @expose('/repo', methods=['POST'])
    @has_dag_access(can_dag_edit=True)
    @provide_session
    def repo_base(self, session=None, path=None):
        return self._git_repo(session, path)

    @expose('/repo/<path:path>', methods=['GET', 'HEAD', 'POST'])
    @has_dag_access(can_dag_edit=True)
    @provide_session
    def repo(self, session=None, path=None):
        return self._git_repo(session, path)

    @expose('/files/<path:path>', methods=['POST'])
    @has_dag_access(can_dag_edit=True)
    @provide_session
    def save(self, session=None, path=None):
        return self._save(session, path)

    @expose('/files/<path:path>', methods=['GET'])
    @has_dag_access(can_dag_edit=True)
    @provide_session
    def load(self, session=None, path=None):
        return self._load(session, path)

    def _render(self, template, *args, **kargs):
        return self.render_template(template + '_appbuilder.html',
                                    airflow_refresh='Airflow.refresh',
                                    log_list='LogModelView.list',
                                    *args, **kargs)


appbuilder_code_editor_view = AppBuilderCodeEditorView()
appbuilder_view = {
    'category': MENU_CATEGORY,
    'name': MENU_LABEL,
    'view': appbuilder_code_editor_view
}
